"""
Evolves the NN using the NEAT algorithm and SUMO with traci extension.
"""
import random
import sys

import visualize  # local file to generate graphs
from conf.settings_traci import TRACI_PATH

sys.path.insert(0, TRACI_PATH)

import neat
import pickle
from typing import Iterable, Tuple, List, Dict

from sumolib.net import Node
from traci._vehicle import VehicleDomain

from conf import settings

import traci

from RouteData import RouteNetwork
from DumbTraci import DumbTraci

# Network
route_net = None


def eval_genome(genome: neat.DefaultGenome, config: neat.Config) -> float:
    """ The fitness function, returns fitness. """
    # Create a new NN with this genome.
    net = neat.nn.FeedForwardNetwork.create(genome, config)

    # Save the fitness for each run.
    fitnesses: List[float] = []

    for runs in range(settings.TEST_RUNS_PER_NET):
        # Simulate the NN, which returns a time.
        # The simulator is not a separate class because of the traci connection it needs.
        fitnesses.append(run_simulator(net, *_create_start_dist(route_net)))

    # The fitness is the average running time
    return sum(fitnesses) / len(fitnesses)


def eval_genomes(genomes: Iterable[Tuple[int, neat.DefaultGenome]], config: neat.Config) -> None:
    """" Pass a genome to eval_genome instead of a list. """
    for _, genome in genomes:
        genome.fitness = eval_genome(genome, config)


def _create_start_dist(net):
    """" Generate a start distribution, returns a dictionary mapping start nodes to an integer amount of cars
     and an ordered version in which each start node corresponds to an index of the List. """
    start: Dict[Node, int] = {}
    start_ordered: List[int] = []
    # net is a sumo network, which contains nodes
    for node in net.net.getNodes():
        # for all start nodes, so not the goal node
        if node.getID() in ('amsterdam',):
            continue
        start[node] = random.randint(20, 25)
        # start[node] = 10
        start_ordered.append(start[node])

    return start, start_ordered


def run_simulator(net: neat.nn.FeedForwardNetwork, start, start_ordered) -> float:
    """" Run the driving simulator with given sumo network and start distribution.
     start_ordered is needed because the network input needs to be ordered. """

    # total travel time
    time: float = 0

    # Total amount of cars randomly generated.
    total_cars: int = sum(start.values())

    # Compute network output
    distribution: List[float] = net.activate(start_ordered)

    # distribution contains for each route the fraction of cars at the start node for that route that is going to be
    # sent over that route, relatively.
    # Relatively means that if the two outgoing routes have 0.75 each, then each route gets 50% of the cars.

    # Run simulation according to the distribution

    # Currently we assume the distribution matrix contains
    # only a car distribution at t=0

    # total sent cars
    total_sent = 0

    # weird traci api
    vehicles: VehicleDomain = traci.vehicle

    # Calculate sum of all distributions for every node, in order to make the fractions relative
    total: Dict[Node, float] = {}
    for node in route_net.net.getNodes():
        total[node] = 0
        for r, route in enumerate(route_net.routes):
            if route.from_node.getID() == node.getID():
                total[node] += max(min(distribution[r], 1), 0)

    # Add the cars at initial positions
    for r, route in enumerate(route_net.routes):
        # find the number of cars to place at this route's start node
        nr_cars = start.get(route.from_node, None)
        if nr_cars is None:
            continue

        # distribution[0][r] / total[from] is a percentage of nr_cars
        # force the fraction between 0 and 1, so we send not more cars than there are or send negative cars
        # then divide by the total of fractions for that start node to make them relative
        send_cars = 0 if total[route.from_node] == 0 else int(
            max(min(distribution[r], 1), 0) / total[route.from_node] * nr_cars)
        total_sent += send_cars

        # add vehicles to traci
        for c in range(send_cars):
            vehicles.add(f'veh{r}_{c}', route.id, typeID='Car')

    # First run a step in order to get the car driving
    # otherwise the check for cars left fails immediately
    traci.simulationStep()

    # while there are cars driving
    while vehicles.getIDList():
        traci.simulationStep()
        time += 1

    # Assume a car takes around x steps to reach the goal,
    # and all leftover cars depart after this time. Sometimes because of rounding a few cars may be left behind.
    time += (total_cars - total_sent) * 100

    # print("\ntotal time steps taken + penalty: ")
    # print(time)
    return -time


def traci_setup(gui) -> None:
    """ Setup connection with traci. """
    # compose the command line to start SUMO
    if gui:
        sumo_binary = "sumo-gui"
    else:
        sumo_binary = "sumo"
    sumo_cmd = [sumo_binary, "-c", "sumomodel/tetrax/tetrax.sumo.cfg", "--no-step-log", "true"]

    # Start simulation and connect to it
    traci.start(sumo_cmd)

    global route_net
    route_net = RouteNetwork('sumomodel/tetrax/tetrax.net.xml')


def run(config) -> None:
    """" Run the network optimizer. """
    traci_setup(gui=False)

    # Make a new population.
    pop = neat.Population(config)

    # Show output.
    # Stats can also be used to make a graph of the fitness
    stats = neat.StatisticsReporter()
    pop.add_reporter(stats)
    pop.add_reporter(neat.StdOutReporter(show_species_detail=True))

    # We cannot use a ParallelEvaluator here because of the traci connection
    winner = pop.run(eval_genomes, settings.NUMBER_EVOLUTIONS)

    traci.close()

    # Save the winner as binary file, to possibly use later.
    with open(settings.BASE_DIR.joinpath('out/winner-feedforward'), 'wb') as f:
        pickle.dump(winner, f)

    # generate a graph of the winner network, and start it up
    visualize.draw_net(config_obj, winner, view=True)
    # Plot the fitness
    visualize.plot_stats(stats, ylog=False, view=True)

    # Create the network from the genome
    winner_network = neat.nn.FeedForwardNetwork.create(winner, config)
    review_winner(winner_network, False)


def review_winner(winner_network, gui) -> None:
    """ Test the winner network visually by running it in the sumo gui. """

    traci_setup(gui=gui)

    dumb_traci = DumbTraci(gui=True, net=route_net)

    for _ in range(10):
        # Create a random distribution
        start_dist = _create_start_dist(route_net)
        print(start_dist[1])
        # print("\nSmart network took steps: ")
        print(-run_simulator(winner_network, *start_dist))

        # print("\nDumb Traci took steps: ")
        print(dumb_traci.run(*start_dist))
        print("\n")


def run_or_debug() -> None:
    """ Choose to evolute a network or to run the existing one, depending on if you 
    hit the debug or run button in PyCharm. """
    # check for debugging to choose what to do
    try:
        import pydevd

        DEBUGGING = True
    except ImportError:
        DEBUGGING = False

    if not DEBUGGING:
        run(config_obj) # evolute a new network
    else:
        # run the last winner network
        with open(settings.BASE_DIR.joinpath('out/winner-feedforward'), 'rb') as f:
            winner = pickle.load(f)
        winner_net = neat.nn.FeedForwardNetwork.create(winner, config_obj)
        review_winner(winner_net, gui=True)

if __name__ == '__main__':
    # Import the config_obj file with settings.
    config_file = str(settings.CONF_DIR.joinpath('config-feedforward'))
    # Initialize the neat config_obj variable.
    config_obj = neat.Config(neat.DefaultGenome, neat.DefaultReproduction, neat.DefaultSpeciesSet,
                             neat.DefaultStagnation, config_file)

    run_or_debug()
